@@ -6,11 +6,11 @@ from __future__ import division, print_function, absolute_import
 
 from scipy import special
 from scipy.special import entr, gammaln as gamln
+from scipy.misc import logsumexp
 
 from numpy import floor, ceil, log, exp, sqrt, log1p, expm1, tanh, cosh, sinh
 
 import numpy as np
-import numpy.random as mtrand
 
 from ._distn_infrastructure import (
         rv_discrete, _lazywhere, _ncx2_pdf, _ncx2_cdf, get_distribution_names)
@@ -31,11 +31,13 @@ class binom_gen(rv_discrete):
 
     `binom` takes ``n`` and ``p`` as shape parameters.
 
+    %(after_notes)s
+
     %(example)s
 
     """
     def _rvs(self, n, p):
-        return mtrand.binomial(n, p, self._size)
+        return self._random_state.binomial(n, p, self._size)
 
     def _argcheck(self, n, p):
         self.b = n
@@ -98,6 +100,8 @@ class bernoulli_gen(binom_gen):
 
     `bernoulli` takes ``p`` as shape parameter.
 
+    %(after_notes)s
+
     %(example)s
 
     """
@@ -145,11 +149,13 @@ class nbinom_gen(rv_discrete):
 
     `nbinom` takes ``n`` and ``p`` as shape parameters.
 
+    %(after_notes)s
+
     %(example)s
 
     """
     def _rvs(self, n, p):
-        return mtrand.negative_binomial(n, p, self._size)
+        return self._random_state.negative_binomial(n, p, self._size)
 
     def _argcheck(self, n, p):
         return (n > 0) & (p >= 0) & (p <= 1)
@@ -202,11 +208,13 @@ class geom_gen(rv_discrete):
 
     `geom` takes ``p`` as shape parameter.
 
+    %(after_notes)s
+
     %(example)s
 
     """
     def _rvs(self, p):
-        return mtrand.geometric(p, size=self._size)
+        return self._random_state.geometric(p, size=self._size)
 
     def _argcheck(self, p):
         return (p <= 1) & (p >= 0)
@@ -215,7 +223,7 @@ class geom_gen(rv_discrete):
         return np.power(1-p, k-1) * p
 
     def _logpmf(self, k, p):
-        return (k-1) * log(1-p) + log(p)
+        return special.xlog1py(k - 1, -p) + log(p)
 
     def _cdf(self, x, p):
         k = floor(x)
@@ -260,6 +268,8 @@ class hypergeom_gen(rv_discrete):
         pmf(k, M, n, N) = choose(n, k) * choose(M - n, N - k) / choose(M, N),
                                        for max(0, N - (M-n)) <= k <= min(n, N)
 
+    %(after_notes)s
+
     Examples
     --------
     >>> from scipy.stats import hypergeom
@@ -295,10 +305,10 @@ class hypergeom_gen(rv_discrete):
 
     """
     def _rvs(self, M, n, N):
-        return mtrand.hypergeometric(n, M-n, N, size=self._size)
+        return self._random_state.hypergeometric(n, M-n, N, size=self._size)
 
     def _argcheck(self, M, n, N):
-        cond = rv_discrete._argcheck(self, M, n, N)
+        cond = (M > 0) & (n >= 0) & (N >= 0)
         cond &= (n <= M) & (N <= M)
         self.a = max(N-(M-n), 0)
         self.b = min(n, N)
@@ -351,6 +361,17 @@ class hypergeom_gen(rv_discrete):
             k2 = np.arange(quant + 1, draw + 1)
             res.append(np.sum(self._pmf(k2, tot, good, draw)))
         return np.asarray(res)
+        
+    def _logsf(self, k, M, n, N):
+        """
+        More precise calculation than log(sf)
+        """
+        res = []
+        for quant, tot, good, draw in zip(k, M, n, N):
+            # Integration over probability mass function using logsumexp
+            k2 = np.arange(quant + 1, draw + 1)
+            res.append(logsumexp(self._logpmf(k2, tot, good, draw)))
+        return np.asarray(res)
 hypergeom = hypergeom_gen(name='hypergeom')
 
 
@@ -370,13 +391,15 @@ class logser_gen(rv_discrete):
 
     `logser` takes ``p`` as shape parameter.
 
+    %(after_notes)s
+
     %(example)s
 
     """
     def _rvs(self, p):
         # looks wrong for p>0.5, too few k=1
         # trying to use generic is worse, no k=1 at all
-        return mtrand.logseries(p, size=self._size)
+        return self._random_state.logseries(p, size=self._size)
 
     def _argcheck(self, p):
         return (p > 0) & (p < 1)
@@ -416,14 +439,21 @@ class poisson_gen(rv_discrete):
 
     `poisson` takes ``mu`` as shape parameter.
 
+    %(after_notes)s
+
     %(example)s
 
     """
+
+    # Override rv_discrete._argcheck to allow mu=0.
+    def _argcheck(self, mu):
+        return mu >= 0
+
     def _rvs(self, mu):
-        return mtrand.poisson(mu, self._size)
+        return self._random_state.poisson(mu, self._size)
 
     def _logpmf(self, k, mu):
-        Pk = k*log(mu)-gamln(k+1) - mu
+        Pk = special.xlogy(k, mu) - gamln(k + 1) - mu
         return Pk
 
     def _pmf(self, k, mu):
@@ -446,9 +476,11 @@ class poisson_gen(rv_discrete):
     def _stats(self, mu):
         var = mu
         tmp = np.asarray(mu)
-        g1 = sqrt(1.0 / tmp)
-        g2 = 1.0 / tmp
+        mu_nonzero = tmp > 0
+        g1 = _lazywhere(mu_nonzero, (tmp,), lambda x: sqrt(1.0/x), np.inf)
+        g2 = _lazywhere(mu_nonzero, (tmp,), lambda x: 1.0/x, np.inf)
         return mu, var, g1, g2
+
 poisson = poisson_gen(name="poisson", longname='A Poisson')
 
 
@@ -467,6 +499,8 @@ class planck_gen(rv_discrete):
 
     `planck` takes ``lambda_`` as shape parameter.
 
+    %(after_notes)s
+
     %(example)s
 
     """
@@ -525,6 +559,8 @@ class boltzmann_gen(rv_discrete):
 
     `boltzmann` takes ``lambda_`` and ``N`` as shape parameters.
 
+    %(after_notes)s
+
     %(example)s
 
     """
@@ -577,6 +613,8 @@ class randint_gen(rv_discrete):
     Note the difference to the numpy ``random_integers`` which
     returns integers on a *closed* interval ``[low, high]``.
 
+    %(after_notes)s
+
     %(example)s
 
     """
@@ -613,7 +651,7 @@ class randint_gen(rv_discrete):
 
         If ``high`` is ``None``, then range is >=0  and < low
         """
-        return mtrand.randint(low, high, self._size)
+        return self._random_state.randint(low, high, self._size)
 
     def _entropy(self, low, high):
         return log(high - low)
@@ -637,11 +675,13 @@ class zipf_gen(rv_discrete):
 
     `zipf` takes ``a`` as shape parameter.
 
+    %(after_notes)s
+
     %(example)s
 
     """
     def _rvs(self, a):
-        return mtrand.zipf(a, size=self._size)
+        return self._random_state.zipf(a, size=self._size)
 
     def _argcheck(self, a):
         return a > 1
@@ -673,6 +713,8 @@ class dlaplace_gen(rv_discrete):
 
     `dlaplace` takes ``a`` as shape parameter.
 
+    %(after_notes)s
+
     %(example)s
 
     """
@@ -727,12 +769,15 @@ class skellam_gen(rv_discrete):
 
     `skellam` takes ``mu1`` and ``mu2`` as shape parameters.
 
+    %(after_notes)s
+
     %(example)s
 
     """
     def _rvs(self, mu1, mu2):
         n = self._size
-        return mtrand.poisson(mu1, n) - mtrand.poisson(mu2, n)
+        return (self._random_state.poisson(mu1, n) -
+                self._random_state.poisson(mu2, n))
 
     def _pmf(self, x, mu1, mu2):
         px = np.where(x < 0,
